import os
import sys

import argparse
import colorama
from pytube import Playlist

try:
    import error
    import font
    from ytam import Downloader
except ModuleNotFoundError:
    import ytam.error as error
    import ytam.font as font
    from ytam.ytam import Downloader



def check_positive(value):
    ivalue = int(value)
    if ivalue <= 0:
        raise argparse.ArgumentTypeError(f"{value} is an invalid positive int value")
    return ivalue

def is_affirmative(string):
    string = string.strip().lower()
    string = string.split(" ")[0]
    truthy = ["y", "yes", "true", "1", "t"]

    return string in truthy


def parse_args(args):
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "url",
        metavar="URL",
        type=str,
        help="the target URL of the playlist to download",
    )
    parser.add_argument(
        "-t",
        "--titles",
        help="a plain text file containing the desired names of the songs in the playlist (each on a new line)",
    )
    parser.add_argument(
        "-d",
        "--directory",
        help="the download directory (defaults to 'music' -  a subdirectory of the current directory)",
    )
    parser.add_argument(
        "-s",
        "--start",
        type=check_positive,
        help="from which position in the playlist to start downloading",
    )
    parser.add_argument(
        "-e",
        "--end",
        type=check_positive,
        help="position in the playlist of the last song to be downloaded",
    )
    parser.add_argument(
        "-A",
        "--album",
        type=str,
        help="the name of the album that the songs in the playlist belongs to (defaults to playlist title)",
    )
    parser.add_argument(
        "-a",
        "--artist",
        type=str,
        help="the name of the artist that performed the songs in the playlist (defaults to Unknown)",
    )
    parser.add_argument(
        "-i",
        "--image",
        type=str,
        help="the path to the image to be used as the album cover (defaults to using the thumbnail of the first video in the playlist). Only works when -A flag is set",
    )
    return parser.parse_args(args)

def main():
    args = parse_args(sys.argv[1:])
    print("Initialising.")
    colorama.init()
    urls = Playlist(args.url)
    playlist_title = urls.title()

    start = 0 if args.start is None else args.start - 1
    end = len(urls) if args.end is None else args.end
    album = playlist_title if args.album is None else args.album
    directory = "music/" if args.directory is None else args.directory
    artist = "Unknown" if args.artist is None else args.artist
    is_album = False if args.album is None else True 

    try:
        if start >= len(urls):
            raise error.InvalidPlaylistIndexError(start, playlist_title)
        if end < start:
            raise error.IndicesOutOfOrderError()

        downloading_message = f"Downloading songs {font.apply('gb', start+1)} - {font.apply('gb', end)} from playlist {font.apply('gb', playlist_title)}"
        text_len = len("Downloading songs ") + len(str(start)) + len(" - ") + len(str(end)) + len(" from playlist ") + len(playlist_title) 
        print(downloading_message, f"\n{font.apply('gb', '─'*text_len)}")
        d = Downloader(urls[start:end], album, directory, artist, is_album, args.titles, args.image)
        d.start = start

        retry = True
        while retry:
            d.download()
            print(f"{font.apply('gb', '─'*text_len)}")
            print(f"{d.successful}/{len(urls[start:end])} downloaded successfully.\n")
            if len(d.retry_urls) > 0:
                d.set_retries()
                user = input(f"Retry {font.apply('fb', str(len(d.urls)) + ' failed')} downloads? Y/N ")
                if not is_affirmative(user):
                    retry = False
                else:
                    print("\nRetrying.")
                    print(f"{font.apply('gb', '─'*len('Retrying.'))}")
            else:
                retry = False

        for image in d.images:
            os.remove(image)
        d.images = []

    except (
        error.InvalidPlaylistIndexError,
        error.IndicesOutOfOrderError,
        error.TitlesNotFoundError,
        error.BadTitleFormatError,
    ) as e:
        print(f"Error: {e.message}")

if __name__ == "__main__":
    main()


import requests

from pytube import YouTube, Playlist
from mutagen.mp4 import MP4, MP4Cover

try:
    import font
    from title import TitleGenerator
except ModuleNotFoundError:
    import ytam.font as font
    from ytam.title import TitleGenerator


def make_safe_filename(string):
    def safe_char(c):
        if c.isalnum():
            return c
        else:
            return "_"

    return "".join(safe_char(c) for c in string).rstrip("_")


class Downloader:
    is_album = None
    album_image_set = False
    urls = None
    album = None
    cur_video = None
    image_filepath = None
    metadata_filepath = None
    successful = 0
    cur_song = 1
    successful_filepaths = []
    retry_urls = []

    start = None

    def __init__(self, urls, album, outdir, artist, is_album, metadata, image_filepath):
        self.urls = urls
        self.album = album
        self.is_album = is_album
        self.outdir = outdir
        self.artist = artist
        self.metadata_filepath = metadata
        self.image_filepath = image_filepath
        self.images = []


    def progress_function(self, chunk, file_handle, bytes_remaining):
        title = self.cur_video.title
        size = self.cur_video.filesize
        p = ((size - bytes_remaining) * 100.0) / size
        progress = (
            f"Downloading song {font.apply('gb',str(self.cur_song))+' - '+font.apply('gb', title)} - [{p:.2f}%]"
            if p < 100
            else f"Downloading song {font.apply('gb',str(self.cur_song))+' - '+font.apply('gb', title)} - {font.apply('bl', '[Done]          ')}"
        )

        end = "\n" if p == 100 else "\r"

        print(progress, end=end, flush=True)

    def apply_metadata(
        self, track_num, total, path, album, title, artist, image_filename
    ):
        song = MP4(path)
        song["\xa9alb"] = album
        song["\xa9nam"] = title
        song["\xa9ART"] = artist
        song["trkn"] = [(track_num, total)]

        with open(image_filename, "rb") as f:
            song["covr"] = [MP4Cover(f.read(), imageformat=MP4Cover.FORMAT_JPEG)]
        song.save()

    @staticmethod
    def download_image(url, index, outdir):
        thumbnail_image = requests.get(url)
        filename = outdir + f"album_art_{str(index)}.jpg"
        with open(filename, "wb",) as f:
            f.write(thumbnail_image.content)
        return filename

    def download(self):
        metadata = None
        if self.metadata_filepath is not None:
            tg = TitleGenerator(self.metadata_filepath, self.artist)
            tg.make_titles()
            metadata = tg.get_titles()

        for num, url in enumerate(self.urls):
            yt = None
            self.cur_song = num+self.start+1
            try:
                yt = YouTube(url)
            except:
                self.retry_urls.append(url)
                print(
                    f"Downloading song {font.apply('gb', str(self.cur_song))} - {font.apply('bf', '[Failed]         ')}\n"
                )
                continue

            path = None
            try:
                yt.register_on_progress_callback(self.progress_function)
                self.cur_video = (
                    yt.streams.filter(type="audio", subtype="mp4")
                    .order_by("abr")
                    .desc()
                    .first()
                )
                path = self.cur_video.download(
                    output_path=self.outdir,
                    filename=make_safe_filename(self.cur_video.title),
                )
                self.successful_filepaths.append(path)
                self.successful += 1
            except Exception as e:
                self.retry_urls.append(url)
                print(
                    f"Downloading song {font.apply('gb',str(self.cur_song))+' - '+font.apply('gb', self.cur_video.title)} - {font.apply('bf', '[Failed]         ')}\n"
                )

                continue

            if self.is_album:
                if self.image_filepath is None:
                    if not self.album_image_set:
                        image_path = Downloader.download_image(
                            yt.thumbnail_url, num, self.outdir
                        )
                        self.images.append(image_path)
                        self.image_filepath = image_path
                        self.album_image_set = True
            else:
                image_path = Downloader.download_image(yt.thumbnail_url, num, self.outdir)
                self.images.append(image_path)
                self.image_filepath = image_path

            track_title = None
            track_artist = None
            if metadata is not None:
                t = metadata[num]
                track_title = t.title if not t.unused else self.cur_video.title
                track_artist = t.artist if not t.unused else self.artist
            else:
                track_title = self.cur_video.title
                track_artist = self.artist

            try:
                self.apply_metadata(
                    num + 1,
                    len(self.urls),
                    path,
                    self.album,
                    track_title,
                    track_artist,
                    self.image_filepath,
                )
                print(f"└── Applying metadata - {font.apply('bl', '[Done]')}\n")

            except:
                print(f"└── Applying metadata - {font.apply('bf', '[Failed]')}\n")


    def set_retries(self):
        self.album_image_set = False
        self.urls = self.retry_urls
        self.retry_urls = []


# if __name__ == "__main__":
#     args = parse_args(sys.argv[1:])
#     print("Initialising.")
#     colorama.init()
#     urls = Playlist(args.url)
#     playlist_title = urls.title()

#     start = 0 if args.start is None else args.start - 1
#     end = len(urls) if args.end is None else args.end
#     album = playlist_title if args.album is None else args.album
#     directory = "music/" if args.directory is None else args.directory
#     artist = "Unknown" if args.artist is None else args.artist
#     is_album = False if args.album is None else True 

#     try:
#         if start >= len(urls):
#             raise error.InvalidPlaylistIndexError(start, playlist_title)
#         if end < start:
#             raise error.IndicesOutOfOrderError()

#         downloading_message = f"Downloading songs {font.apply('gb', start+1)} - {font.apply('gb', end)} from playlist {font.apply('gb', playlist_title)}"
#         text_len = len("Downloading songs ") + len(str(start)) + len(" - ") + len(str(end)) + len(" from playlist ") + len(playlist_title) 
#         print(downloading_message, f"\n{font.apply('gb', '─'*text_len)}")
#         d = Downloader(urls[start:end], album, directory, artist, is_album, args.titles, args.image)
#         d.start = start

#         retry = True
#         while retry:
#             d.download()
#             print(f"{font.apply('gb', '─'*text_len)}")
#             print(f"{d.successful}/{len(urls[start:end])} downloaded successfully.\n")
#             if len(d.retry_urls) > 0:
#                 d.set_retries()
#                 user = input(f"Retry {font.apply('fb', str(len(d.urls)) + ' failed')} downloads? Y/N ")
#                 if not is_affirmative(user):
#                     retry = False
#                 else:
#                     print("\nRetrying.")
#                     print(f"{font.apply('gb', '─'*len('Retrying.'))}")
#             else:
#                 retry = False

#         for image in d.images:
#             os.remove(image)
#         d.images = []

#     except (
#         error.InvalidPlaylistIndexError,
#         error.IndicesOutOfOrderError,
#         error.TitlesNotFoundError,
#         error.BadTitleFormatError,
#     ) as e:
#         print(f"Error: {e.message}")


import os
try:
    import error
except ModuleNotFoundError:
    import ytam.error as error

class Title:
    def __init__(self, index, title, artist):
        self.unused = False
        self.index = index
        self.title = title
        self.artist = artist


class TitleGenerator:
    def __init__(self, filename, artist):
        self.filename = filename
        self.artist = artist
        self.titles = []

    def check_line(self, line, line_num):
        if not (len(line) > 0 and len(line) < 3):
            msg = "wrong number of fields - only <title>, <artist> allowed"
            raise error.BadTitleFormatError(self.filename, line_num + 1, msg)

    def make_titles(self):
        try:
            with open(self.filename, "r") as f:
                lines = f.readlines()

                for i, line in enumerate(lines):
                    line = line.strip()
                    line = line.split(",")
                    self.check_line(line, i)

                    t = Title(i, None, None)
                    if line[0] == "":
                        t.unused = True

                    if len(line) == 1:
                        t.title = line[0].strip()
                        t.artist = self.artist.strip()
                    elif len(line) == 2:
                        t.title = line[0].strip()
                        t.artist = line[1].strip()
                    self.titles.append(t)
        except FileNotFoundError:
            raise error.TitlesNotFoundError(self.filename)

    def get_titles(self):
        return self.titles


if __name__ == "__main__":
    t = TitleGenerator("./titles.txt", "Cool Artist")
    t.make_titles()
    ts = t.get_titles()

    for t in ts:
        print(
            f"{t.index}: {t.title}, {t.artist} - {('un' if t.unused else '' )+ 'used'}"
        )



import setuptools
from setuptools import find_namespace_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ytam", #change to ytam
    version="0.0.15",
    author="jayathungek",
    author_email="jayathunge.work@gmail.com",
    description="A commandline utility that enables the creation of albums from Youtube playlists.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pypa/sampleproject",
    packages=["ytam"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    entry_points = {
        'console_scripts': ['ytam=ytam.cmd:main'],
    },
    install_requires=[
        "certifi",
        "chardet",
        "colorama",
        "idna",
        "mutagen",
        "requests",
        "urllib3"
    ]
)


HEADER = "\033[95m"
OKBLUE = "\033[94m"
OKGREEN = "\033[92m"
WARNING = "\033[93m"
FAIL = "\033[91m"
ENDC = "\033[0m"
BOLD = "\033[1m"
UNDERLINE = "\033[4m"


def header(text):
    return f"{HEADER}{text}{ENDC}"


def ok_blue(text):
    return f"{OKBLUE}{text}{ENDC}"


def ok_green(text):
    return f"{OKGREEN}{text}{ENDC}"


def warning(text):
    return f"{WARNING}{text}{ENDC}"


def fail(text):
    return f"{FAIL}{text}{ENDC}"


def bold(text):
    return f"{BOLD}{text}{ENDC}"


def underline(text):
    return f"{UNDERLINE}{text}{ENDC}"


lookup = {
    "h": header,
    "l": ok_blue,
    "g": ok_green,
    "w": warning,
    "f": fail,
    "b": bold,
    "u": underline,
}


def apply(params, text):
    formatted_text = text
    for style in params:
        formatted_text = lookup[style](formatted_text)

    return formatted_text



class Error(Exception):
    """Base class for exceptions in this module."""

    pass


class EmptyUrlFieldError(Error):
    def __init__(self):
        self.message = "URL field cannot be empty."


class InvalidUrlError(Error):
    def __init__(self, url):
        self.message = f"URL {url} is not a valid YouTube playlist link."


class InvalidPlaylistError(Error):
    def __init__(self):
        self.message = "Playlist is invalid."


class TitlesNotFoundError(Error):
    def __init__(self, filename):
        self.message = f"TIL file {filename} not found."


class BadTitleFormatError(Error):
    def __init__(self, filename, line, msg):
        self.message = f"Bad formatting on line {line} of {filename}: {msg}."


class InvalidPlaylistIndexError(Error):
    def __init__(self, index, title):
        self.message = f"Start index {index} is out of range for playlist {title}."


class IndicesOutOfOrderError(Error):
    def __init__(self):
        self.message = f"End index must be greater that start index."


class InvalidFieldError(Error):
    def __init__(self, field, msg):
        self.message = f"Invalid argument for field {field} - {msg}."


class InvalidPathError(Error):
    def __init__(self, path):
        self.message = f"DIR {path} is not a valid directory."


# bad url
# url not for playlist
# title file not found
# invalid line in title file
# directory not found
